#include <stdlib.h>
#include <netdb.h>
#include <sys/select.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>

static struct sockaddr_storage pgStatAddr;

typedef int bool;

void
socket_test_send_recv(void)
{

    int pgStatSock = -1;
    int pgStatSock2 = -1;
	socklen_t alen;
	struct addrinfo *addrs = NULL,
			   *addr,
				hints;
	int			ret;
	fd_set		rset;
	struct timeval tv;
	char		test_byte;
	int			sel_res;
	bool retry = 1;

#define TESTBYTEVAL ((char) 199)

	/*
	 * Create the UDP socket for sending and receiving statistic messages
	 */
	hints.ai_flags = AI_PASSIVE;
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_DGRAM;
	hints.ai_protocol = 0;
	hints.ai_addrlen = 0;
	hints.ai_addr = NULL;
	hints.ai_canonname = NULL;
	hints.ai_next = NULL;
	ret = getaddrinfo("localhost", NULL, &hints, &addrs);
	if (ret || !addrs) {
		printf("could not resolve \"localhost\": %s\n", gai_strerror(ret));
	}
	else {

		/*
		 * Loop through the results to try and find a working address.
		 */
		for (addr = addrs; addr; addr = addr->ai_next)
		{
			/*
			 * Create the socket.
			 */
			if ((pgStatSock = socket(addr->ai_family, SOCK_DGRAM, 0)) == -1)
			{
				printf("could not create socket for statistics collector\n");
				continue;
			}
	
			/*
			 * Bind it to a kernel assigned port on localhost and get the assigned
			 * port via getsockname().
			 */
			if (bind(pgStatSock, addr->ai_addr, addr->ai_addrlen) < 0)
			{
				printf("could not bind socket for statistics collector\n");
				closesocket(pgStatSock);
				pgStatSock = -1;
				continue;
			}
	
			alen = sizeof(pgStatAddr);
			if (getsockname(pgStatSock, (struct sockaddr *) & pgStatAddr, &alen) < 0)
			{
				printf("could not get address of socket for statistics collector\n");
				closesocket(pgStatSock);
				pgStatSock = -1;
				continue;
			}
	
			/*
			 * Open a second socket to do the send as we apparently can't
			 * send through the socket that we use to receive the data.
			 */
			if ((pgStatSock2 = socket(addr->ai_family, SOCK_DGRAM, 0)) == -1)
			{
				printf("could not create socket for statistics collector\n");
				continue;
			}
			if (connect(pgStatSock2, (struct sockaddr *) & pgStatAddr, alen) < 0)
			{
				printf("could not connect socket for statistics collector\n");
				closesocket(pgStatSock2);
				pgStatSock2 = -1;
				continue;
			}
	
			/*
			 * Try to send and receive a one-byte test message on the socket. This
			 * is to catch situations where the socket can be created but will not
			 * actually pass data (for instance, because kernel packet filtering
			 * rules prevent it).
			 */
			test_byte = TESTBYTEVAL;
	
			while (retry) {
				if (send(pgStatSock2, &test_byte, 1, 0) != 1) {
					/* if not interrupted, bail out */
					if (errno != EINTR) {
						retry = 0;
						printf("could not send test message on socket for statistics collector\n");
						closesocket(pgStatSock2);
						pgStatSock2 = -1;
						continue;
					}
				}
				else {
					retry = 0;
				}
			}
	
			/*
			 * There could possibly be a little delay before the message can be
			 * received.  We arbitrarily allow up to half a second before deciding
			 * it's broken.
			 */
			for (;;)				/* need a loop to handle EINTR */
			{
				FD_ZERO(&rset);
				FD_SET(pgStatSock, &rset);
	
				tv.tv_sec = 0;
				tv.tv_usec = 500000;
				sel_res = select(pgStatSock + 1, &rset, NULL, NULL, &tv);
				if (sel_res >= 0 || errno != EINTR)
					break;
			}
			if (sel_res < 0)
			{
				printf("select() failed in statistics collector\n");
				closesocket(pgStatSock);
				pgStatSock = -1;
				continue;
			}
			if (sel_res == 0 || !FD_ISSET(pgStatSock, &rset))
			{
				/*
				 * This is the case we actually think is likely, so take pains to
				 * give a specific message for it.
				 *
				 * errno will not be set meaningfully here, so don't use it.
				 */
				printf("test message did not get through on socket for statistics collector\n");
				closesocket(pgStatSock);
				pgStatSock = -1;
				continue;
			}
	
			test_byte++;			/* just make sure variable is changed */
	
			retry = 1;
			while (retry) {
				if (recv(pgStatSock, &test_byte, 1, 0) != 1) {
					/* if not interrupted, bail out */
					if (errno != EINTR) {
						retry = 0;
						printf("Errno: %i\n", errno);
						printf("String error: %s\n", strerror(errno));
						printf("could not receive test message on socket for statistics collector\n");
						closesocket(pgStatSock);
						pgStatSock = -1;
						continue;
					}
				}
				else {
					retry = 0;
				}
			}
	
			if (test_byte != TESTBYTEVAL)	/* strictly paranoia ... */
			{
				printf("incorrect test message transmission on socket for statistics collector\n");
				closesocket(pgStatSock);
				pgStatSock = -1;
				continue;
			}
	
			/* If we get here, we have a working socket */
			printf("We have a working socket!\n");
			
			if (pgStatSock > -1) {
				closesocket(pgStatSock);
			}
			if (pgStatSock2 > -1) {
				closesocket(pgStatSock2);
			}
			break;
		}
	}

	if (addrs != NULL)
		freeaddrinfo(addrs);

}

int main(int argc, char **argv) {
	
	socket_test_send_recv();

	return 0;
	
}
